///////////////////////////////////////////////////////////////////////////////
// Plasma Fence 1.0 - A library to detect illegal memory accesses
// Copyright (c) 2014 Hadrien Nilsson - psydk.org
// This code is distributed under the terms of the
// GNU LESSER GENERAL PUBLIC LICENSE. See LICENSE file for the full license.
///////////////////////////////////////////////////////////////////////////////

#include <stddef.h>
#include <pthread.h>
#include <stdint.h>
#include <unistd.h>
#include <sys/mman.h>
#include <string.h>
#include <signal.h>
#include <errno.h>

///////////////////////////////////////////////////////////////////////////////
// User options

// Small allocations are recycled to avoid slowing down execution.
// Set to true to protect the data part of the small buffer after it
// was freed. It adds extra verifications but slowdowns execution.
// false: best speed (default)
static const bool k_protectFreeSmallBlocks = false;

// If a memcpy fails it can be hard to determine the adresses, size and failing
// operation (read or write) unless we are using the debug version of the libc.
// We can makes things easier to understand in our custom memcpy.
// undef: best speed (default)
//#define OVERRIDE_MEMCPY 1

///////////////////////////////////////////////////////////////////////////////
// Internal stuff

static const uint32_t k_magic1 = 0xaaaaaaaa;
static const uint32_t k_magic2 = 0xbbbbbbbb;
static const int      k_default_alignment = 16; // 8 for 32 bits
static const uint8_t  k_default_fill_byte = 0xcd;
static const uint8_t  k_padding_byte = 0xff;

// We could use constants instead, because 4k pages are very common
static int    g_page_size = 0;
static int    g_page_shift = 0;
static size_t g_page_mask = 0;

static bool   g_initialized = false;

// This is a list of small blocks, each one is made of two pages:
// The first page contains the block header and the user data,
// the second page is the fence page. Most programs perform a lot
// of small allocations, thus we keep them in that list and recycle
// when needed to keep things fast.
static struct SmallBlock* g_sb_head = nullptr;
static pthread_mutex_t g_mutex; // That's the mutex to protect the list

///////////////////////////////////////////////////////////////////////////////
static inline int compute_log2(int val)
{
	return __builtin_ctz(val);
}

static inline size_t round_to_next_page(size_t val)
{
	size_t next_page = (val + g_page_size-1);
	next_page >>= g_page_shift;
	next_page <<= g_page_shift;
	return next_page;
}

static inline size_t round_to_this_page(size_t val)
{
	val >>= g_page_shift;
	val <<= g_page_shift;
	return val;
}

///////////////////////////////////////////////////////////////////////////////
// All allocations have this header. I kept the size to 16 because I saw that
// some programs expected it. For example: multiple allocations of 4080 bytes.
// It's tough to keep such a small header, so for now allocations are limited
// to 0xffffffff bytes.
struct BlockHeader
{
	uint32_t magic1;  // To detect damages
	uint32_t size;    // User size as requested in malloc() and friends
	uint16_t pcbh;    // Page count before header (complicated alignments)
	uint16_t pcaf;    // Page count after fence page (complicated alignments)
	uint32_t magic2;  // To detect damages too
	uint8_t  data[0]; // User data starts here

	// User data is kept as close as possible to the fence page, but
	// because of alignment constraints, padding bytes occurs. Those
	// padding bytes are filled with special bytes and checked for
	// damaged every time the block is given as an argument to pfence.
	inline size_t get_padding_size() const
	{
		size_t size_end = reinterpret_cast<size_t>(data + size);
		size_t next_page = round_to_next_page(size_end);
		size_t padding_size = next_page - size_end;
		return padding_size;
	}
};

// A small block is made of two pages and can hold one page of data
// minus the BlockHeader size not counting the user data array
struct SmallBlock
{
	SmallBlock* next;
};

///////////////////////////////////////////////////////////////////////////////
// A very primitive printing function that does not allocate anything.
static void pfence_print(const char* msg,
						 const char* msg2 = nullptr,
						 const char* msg3 = nullptr)
{
	write(1, msg, strlen(msg));
	if( msg2 )
		write(1, msg2, strlen(msg2));
	if( msg3 )
		write(1, msg3, strlen(msg3));
}

///////////////////////////////////////////////////////////////////////////////
// Prints some message and stops execution with a signal that can be caught
// by a debugger.
static void pfence_abort(const char* msg,
						 const char* msg2 = nullptr)
{
	const char premsg[] = "[pfence] abort! ";
	int len = sizeof(premsg)-1;
	write(2, premsg, len);
	write(2, msg, strlen(msg));
	if( msg2 )
	{
		write(2, ": ", 3);
		write(2, msg2, strlen(msg2));
	}
	write(2, "\n", 1);
	kill(getpid(), SIGILL);
}

///////////////////////////////////////////////////////////////////////////////
// Waits for data on stdin. It's to get some time to attach a debugger before
// the attached process crashes.
static void pfence_wait_user(const char* msg,
							 const char* msg2 = nullptr,
							 const char* msg3 = nullptr)
{
	pfence_print(msg, msg2, msg3);
	char buf[8];
	read(0, buf, 1);
}

///////////////////////////////////////////////////////////////////////////////
// The main memory allocation function.
//
// [in] func_name   To give the public function name upon abort
// [in] size        Size in byte of the allocation
// [in] alignment   Alignment constraint of the buffer. It is 8 on 32 bits
//                  systems and 16 on 64 bits system. By default we use 16 but
//                  it increases the probability of getting padding bytes.
// [in] fill_byte   By default we use 0xcd so the user can identify uninitialized
//                  data, but we accept it as an argument for calloc implementation
//
// Returns the address of the user data
///////////////////////////////////////////////////////////////////////////////
static void* pfence_malloc(const char* func_name,
						   size_t size, int alignment,
						   uint8_t fill_byte = k_default_fill_byte)
{
	if( alignment < k_default_alignment )
	{
		alignment = k_default_alignment;
	}
	int adjust = alignment;
	if( adjust > g_page_size )
	{
		adjust = g_page_size;
	}
	size_t rounded_size = ((size + adjust-1) / adjust)*adjust;

	if( rounded_size > 0xffffffffu )
	{
		pfence_abort(func_name, "size to high");
	}

	BlockHeader* bh = nullptr;

	const size_t sb_max_data = g_page_size - sizeof(BlockHeader);
	if( rounded_size <= sb_max_data )
	{
		// Check in the free small block list
		pthread_mutex_lock(&g_mutex);
		if( g_sb_head )
		{
			SmallBlock* block = g_sb_head;

			if( k_protectFreeSmallBlocks )
			{
				// Make the page writable again
				if( mprotect(block, g_page_size, PROT_READ|PROT_WRITE) != 0 )
				{
					pfence_abort(func_name, "sb mprotect existing failed");
				}
			}            
			g_sb_head = block->next;

			size_t bh_page = reinterpret_cast<size_t>(block);
			size_t bh_start = bh_page + g_page_size
					- rounded_size
					- sizeof(BlockHeader);
			bh = reinterpret_cast<BlockHeader*>(bh_start);
			bh->pcbh = 0;
			bh->pcaf = 0;
		}
		pthread_mutex_unlock(&g_mutex);
	}

	if( !bh )
	{
		// ...[adjust][adjust][header|data|padding]
		//    -------------------------------------
		//    [fencepage][adjust][adjust]...
		size_t paged_size1 = round_to_next_page(sizeof(BlockHeader) + rounded_size);
		size_t paged_size2 = round_to_next_page(alignment);

		size_t paged_size = paged_size1 + paged_size2;
		void* page_addr = mmap(nullptr, paged_size, PROT_NONE,
					   MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
		if( page_addr == MAP_FAILED )
		{
			pfence_abort(func_name, "mmap failed");
		}

		// Find the fence page that will satisfy alignment

		size_t page = reinterpret_cast<size_t>(page_addr);
		size_t data = page + paged_size - g_page_size - rounded_size;
		size_t data2 = (data / alignment) * alignment;

		size_t fence_page_addr = data2 + rounded_size;
		size_t full_fence_size = page + paged_size - fence_page_addr;

		size_t bh_start = data2 - sizeof(BlockHeader);
		size_t bh_page = round_to_this_page(bh_start);
		size_t rw_size = fence_page_addr - bh_page;
		void* bh_page_addr = reinterpret_cast<void*>(bh_page);

		if( mprotect(bh_page_addr, rw_size, PROT_READ|PROT_WRITE) != 0 )
		{
			pfence_abort(func_name, "mprotect failed, try increasing max_map_count:\n"
				"[pfence] $ sudo su\n"
				"[pfence] $ echo 200000 > /proc/sys/vm/max_map_count\n"
				"[pfence] $ exit\n"
				"[pfence] change 200000 for something greater if the problem persists");
		}

		size_t pcbh = (bh_page - page) >> g_page_shift;
		size_t pcaf = (full_fence_size - g_page_size) >> g_page_shift;

		bh = reinterpret_cast<BlockHeader*>(bh_start);
		bh->pcbh = static_cast<uint16_t>(pcbh);
		bh->pcaf = static_cast<uint16_t>(pcaf);
	}

	bh->size = size;
	bh->magic1 = k_magic1;
	bh->magic2 = k_magic2;

	size_t padding_size = bh->get_padding_size();

	// Clean memory with debug bytes and add padding bytes
	memset(bh->data, fill_byte, size);
	memset(bh->data + size, k_padding_byte, padding_size);
	return bh->data;
}

///////////////////////////////////////////////////////////////////////////////
// Gets the memory block owning user data. May abort if we detect a problem in
// the block, like damaged bytes or a plain wrong address.
//
// [in] func_name  Used upon abort to know the public function called
// [in] addr       User data address
//
// Returns the block address
///////////////////////////////////////////////////////////////////////////////
static BlockHeader* pfence_get_block(const char* func_name, void* addr)
{
	uint8_t* addr_bytes = static_cast<uint8_t*>(addr);
	addr_bytes -= sizeof(BlockHeader);
	BlockHeader* bh = reinterpret_cast<BlockHeader*>(addr_bytes);
	if( bh->magic1 != k_magic1 )
	{
		pfence_abort(func_name, "magic1");
	}
	if( bh->magic2 != k_magic2 )
	{
		pfence_abort(func_name, "magic2");
	}
	size_t padding_size = bh->get_padding_size();
	const uint8_t* padding = bh->data + bh->size;
	for(size_t i = 0; i < padding_size; ++i)
	{
		if( padding[i] != k_padding_byte )
		{
			pfence_abort(func_name, "padding damaged");
		}
	}
	return bh;
}

///////////////////////////////////////////////////////////////////////////////
static void pfence_free(void* addr)
{
	if( !addr )
	{
		// The spec says nothing occurs for null pointer
		return;
	}

	BlockHeader* bh = pfence_get_block("free", addr);
	size_t page_addr_int = reinterpret_cast<size_t>(bh);
	page_addr_int = round_to_this_page(page_addr_int);
	void* page = reinterpret_cast<void*>(page_addr_int);

	size_t rounded_size = bh->size + bh->get_padding_size();
	const size_t sb_max_data = g_page_size - sizeof(BlockHeader);
	if( rounded_size <= sb_max_data )
	{
		// Give back the block to the small block list
		pthread_mutex_lock(&g_mutex);
		SmallBlock* next = nullptr;
		if( g_sb_head )
		{
			next = g_sb_head;
		}
		g_sb_head = static_cast<SmallBlock*>(page);
		g_sb_head->next = next;
		
		if( k_protectFreeSmallBlocks )
		{
			if( mprotect(page, g_page_size, PROT_NONE) != 0 )
			{
				pfence_abort("free", "sb mprotect failed");
			}
		}
		pthread_mutex_unlock(&g_mutex);
	}
	else
	{
		// So many lines just because I didn't want to keep track
		// of the total size and the first page address which was
		// not possible without increasing the header size over 16.
		size_t map_begin = reinterpret_cast<size_t>(page);
		map_begin -= bh->pcbh * g_page_size;
		size_t data = reinterpret_cast<size_t>(bh->data);
		data += bh->size;
		size_t map_end = round_to_next_page(data);
		map_end += g_page_size;
		map_end += bh->pcaf * g_page_size;
		size_t mapping_size = map_end - map_begin;
		void* map_begin_addr = reinterpret_cast<void*>(map_begin);
		munmap(map_begin_addr, mapping_size);
	}
}

///////////////////////////////////////////////////////////////////////////////
static void* pfence_realloc(void* old_addr, size_t new_size)
{
	// Get the block first to validate because we want
	// validation to be the first thing to do when
	// entering the function
	size_t old_size = 0;
	if( old_addr )
	{
		BlockHeader* bh = pfence_get_block("realloc", old_addr);
		old_size = bh->size;
	}

	// A real realloc function could optimize reallocation but because in pfence
	// we need to push the user data down to the page, we cannot grow with zero
	// penalty. We just alloc a new block and copy the old content if it exists.
	void* new_addr = pfence_malloc("realloc", new_size, k_default_alignment);

	if( old_addr )
	{
		size_t cpy_size = old_size < new_size ? old_size : new_size;
		memcpy(new_addr, old_addr, cpy_size);
		pfence_free(old_addr);
	}
	return new_addr;
}

///////////////////////////////////////////////////////////////////////////////
// I read so many things on the Web about itoa in the libc that I decided to
// write my own simple and specific version here.
static char* itoa_dec(int val, char* buf, int buf_size)
{
	static const char ascii_digits[] = "0123456789";
	buf += buf_size - 1;
	buf[0] = 0;
	while(val)
	{
		buf--;
		buf[0] = ascii_digits[val % 10];
		val /= 10;
	}
	return buf;
}

///////////////////////////////////////////////////////////////////////////////
// This function is called when the first allocation function is called.
// It is not thread safe but we expect the libc to perform the first allocation
// before main(), when chances of having already multiple threads is very low.
static inline void pfence_init()
{
	if( g_initialized )
	{
		return;
	}
	g_page_size = getpagesize();
	g_page_shift = compute_log2(g_page_size);
	g_page_mask = ~(g_page_size - 1);
	pthread_mutex_init(&g_mutex, nullptr);

	char tmp[20];
	pfence_wait_user("[pfence] init done, it's time to attach a debugger:\n"
		"[pfence] $ sudo gdb -p ",
		itoa_dec(getpid(), tmp, sizeof(tmp)),
		"\n[pfence] press return to continue");

	g_initialized = true;
}

///////////////////////////////////////////////////////////////////////////////
// Just to make easy to debug pfence code, see main() at the bottom of this file.
#ifndef TEST

///////////////////////////////////////////////////////////////////////////////
// Here come the redefined public functions
extern "C" {

///////////////////////////////////////////////////////////////////////////////
void* malloc(size_t size)
{
	pfence_init();
	return pfence_malloc("malloc", size, k_default_alignment);
}

///////////////////////////////////////////////////////////////////////////////
void* calloc(size_t num, size_t size)
{
	pfence_init();
	return pfence_malloc("calloc", num*size, k_default_alignment, 0);
}

///////////////////////////////////////////////////////////////////////////////
int posix_memalign(void** memptr, size_t alignment, size_t size)
{
	// spec says alignment must be multiple of void*
	const int bppa = (sizeof(void*)/4)+1;
	const size_t mask = (1<<bppa) - 1;
	if( alignment & mask )
	{
		return EINVAL;
	}
	// We also check that alignment is a power of two
	if( alignment & (alignment - 1) )
	{
		return EINVAL;
	}

	pfence_init();
	*memptr = pfence_malloc("posix_memalign", size, alignment);
	if( !*memptr )
	{
		return ENOMEM;
	}
	return 0;
}

///////////////////////////////////////////////////////////////////////////////
void* aligned_alloc(size_t alignment, size_t size)
{
	pfence_init();
	return pfence_malloc("aligned_alloc", size, alignment);
}

///////////////////////////////////////////////////////////////////////////////
void* valloc(size_t size)
{
	pfence_init();
	return pfence_malloc("valloc", size, g_page_size);
}

///////////////////////////////////////////////////////////////////////////////
void* pvalloc(size_t size)
{
	pfence_init();
	return pfence_malloc("pvalloc", size, g_page_size);
}

///////////////////////////////////////////////////////////////////////////////
void* memalign(size_t alignment, size_t size)
{
	pfence_init();
	return pfence_malloc("memalign", size, alignment);
}

///////////////////////////////////////////////////////////////////////////////
void free(void* addr)
{
	pfence_init();
	pfence_free(addr);
}

///////////////////////////////////////////////////////////////////////////////
void* realloc(void* old_addr, size_t new_size)
{
	pfence_init();
	return pfence_realloc(old_addr, new_size);
}

///////////////////////////////////////////////////////////////////////////////
size_t malloc_usable_size(void* addr)
{
	pfence_init();
	if( !addr )
		return 0;
	BlockHeader* bh = pfence_get_block("malloc_usable_size", addr);
	return bh->size;
}

///////////////////////////////////////////////////////////////////////////////
#ifdef OVERRIDE_MEMCPY

void* memcpy(void* dst, const void* src, size_t n)
{
	// We cannot be sure dst and src were allocated with our malloc
	// so we do not try to get a possible BlockHeader from them.

	size_t dst_int = reinterpret_cast<size_t>(dst);
	size_t src_int = reinterpret_cast<size_t>(src);

	// At least we try to optimize a bit the copy for 32 bits copies
	if( (dst_int & 0x03) == 0 && (src_int & 0x03) == 0 && (n & 0x03) == 0 )
	{
		uint32_t* dst32 = static_cast<uint32_t*>(dst);
		const uint32_t* src32 = static_cast<const uint32_t*>(src);
		n /= 4;
		for(size_t i = 0; i < n; ++i)
		{
			uint32_t word = src32[i]; // Crash upon read?
			dst32[i] = word; // Crash upon write?
		}
	}
	else
	{
		uint8_t* dst8 = static_cast<uint8_t*>(dst);
		const uint8_t* src8 = static_cast<const uint8_t*>(src);
		for(size_t i = 0; i < n; ++i)
		{
			uint8_t byte = src8[i]; // Crash upon read?
			dst8[i] = byte; // Crash upon write?
		}
	}
	return dst;
}
#endif

///////////////////////////////////////////////////////////////////////////////
}

#else // TEST

// Simple code to debug and test corner cases of pfence.
// Keeps changing during development
int main()
{
	pfence_init();
	char* p = (char*)pfence_malloc("", 16, 16);
	p[17] = 1;
}
#endif
